#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import os
import sys
import traceback
import warnings

from threadpoolctl import threadpool_limits
import MDAnalysis as mda
import maicos

from . import __version__
from . import __all__ as available_modules
from .utils import get_cli_input
from .arg_completion import complete_parser

# Try to use IPython shell for debug
try:
    import IPython
    use_IPython = True
except ImportError:
    import code
    use_IPython = False


class bcolors:
    warning = '\033[93m'
    fail = '\033[91m'
    endc = '\033[0m'


def _warning(message,
             category=UserWarning,
             filename='',
             lineno=-1,
             file=None,
             line=None):
    print("{}Warning: {}{}".format(bcolors.warning, message, bcolors.endc))


warnings.showwarning = _warning


def parse_args():
    """Parse the command line arguments."""

    if '--bash_completion' in sys.argv:
        print(
            os.path.join(os.path.dirname(__file__),
                         "share/maicos-completion.bash"))
        sys.exit(0)

    parser = argparse.ArgumentParser(
        description="Analyse molecular dynamics simulations of "
        "interfacial and confined systems.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("program",
                        type=str,
                        help="Program to start",
                        choices=available_modules)
    parser.add_argument('--debug',
                        action='store_true',
                        help="Run with debug options. Will start an "
                        "interactive Python interpreter at the end of "
                        "the program.")
    parser.add_argument('--version',
                        action='version',
                        version="maicos {}".format(__version__))

    try:
        sys.argv.remove("--debug")
        debug = True
    except ValueError:
        debug = False
        warnings.filterwarnings("ignore")

    try:
        if sys.argv[1] in available_modules:
            selected_module = getattr(maicos, sys.argv[1])
        else:
            parser.parse_args()
    except IndexError:
        parser.parse_args()

    print('\n{}\n'.format(get_cli_input()))
    parser = argparse.ArgumentParser(
        prog="maicos " + sys.argv[1],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("-s",
                        dest="topology",
                        type=str,
                        default="topol.tpr",
                        help="The topolgy file. "
                        "The FORMATs {} are implemented in MDAnalysis."
                        "".format(", ".join(mda._PARSERS.keys())))
    parser.add_argument("-top",
                        dest="topology_format",
                        type=str,
                        default=None,
                        help="Override automatic topology type detection. "
                        "See topology for implemented formats")
    parser.add_argument("-f",
                        dest="trajectory",
                        type=str,
                        default=None,
                        nargs="+",
                        help="A single or multiple trajectory files. "
                        "The FORMATs {} are implemented in MDAnalysis."
                        "".format(", ".join(mda._READERS.keys())))
    parser.add_argument("-traj",
                        dest="trajectory_format",
                        type=str,
                        default=None,
                        help="Override automatic trajectory type detection. "
                        "See trajectory for implemented formats")
    parser.add_argument("-atom_style",
                        dest="atom_style",
                        type=str,
                        default=None,
                        help="Manually set the atom_style information"
                        "(currently only LAMMPS parser)."
                        "E.g. atom_style='id type x y z'.")
    parser.add_argument("-b",
                        dest="begin",
                        type=float,
                        default=0,
                        help="start time (ps) for evaluation.")
    parser.add_argument("-e",
                        dest="end",
                        type=float,
                        default=None,
                        help="end time (ps) for evaluation.")
    parser.add_argument("-dt",
                        dest="dt",
                        type=float,
                        default=0,
                        help="time step (ps) to read analysis frame. "
                        "If `0` take all frames")
    parser.add_argument("-box",
                        dest="box",
                        type=float,
                        default=None,
                        nargs="+",
                        help="Sets the box dimensions x y z [alpha beta gamma]"
                        "(Å). If 'None' dimensions from the trajectory "
                        "will be used.")
    parser.add_argument("-nt",
                        dest="num_threads",
                        type=int,
                        default=0,
                        help="Total number of threads to start (0 is guess)")

    try:
        _configure_parser = getattr(selected_module, "_configure_parser")
        _allow_multiple_atomgroups = getattr(selected_module,
                                             "_allow_multiple_atomgroups")

        if _allow_multiple_atomgroups:
            parser.add_argument(
                '-sel',
                dest='selection',
                type=str,
                default=['all'],
                nargs='+',
                help='Atomgroup(s) for which to perform the analysis.',
            )
        else:
            parser.add_argument(
                '-sel',
                dest='selection',
                type=str,
                default='all',
                help='Atomgroup for which to perform the analysis.',
            )

        _configure_parser(selected_module, parser)
        complete_parser(parser, selected_module)
        args = parser.parse_args(sys.argv[2:])
        args.debug = debug
        args._allow_multiple_atomgroups = _allow_multiple_atomgroups
        args.selected_module = selected_module
    except Exception as e:
        if debug:
            traceback.print_exc()
        else:
            print("{}Error: {}{}".format(bcolors.fail, e, bcolors.endc))

    return args


def main(args, verbose=True):
    """The maicos main function including universe initialization and module running."""

    try:
        with threadpool_limits(limits=args.num_threads):
            if verbose:
                print("Loading trajectory... ", end="")
                sys.stdout.flush()

            # prepare kwargs dictionary for other optional arguments
            ukwargs = {}
            if args.atom_style is not None:
                ukwargs['atom_style'] = args.atom_style

            u = mda.Universe(args.topology,
                             topology_format=args.topology_format,
                             **ukwargs)
            if args.trajectory is not None:
                u.load_new(args.trajectory, format=args.trajectory_format)
            if verbose:
                print("Done!\n")

            if args.box is not None:
                if len(args.box) == 6:
                    u.dimensions = args.box
                if len(args.box) == 3:
                    u.dimensions[:3] = args.box
                    u.dimensions[3:] = [90, 90, 90]
                else:
                    raise IndexError(
                        "The boxdimensions must contain 3 entries for "
                        "the box length and possibly 3 more for the angles.")
            if verbose:
                print("Performing analysis for the following group(s):")

            if type(args.selection) is str:
                args.selection = [args.selection]

            atomgroups = []
            for i, gr in enumerate(args.selection):
                sel = u.select_atoms(gr)
                if verbose:
                    print("{:>15}: {:>10} atoms".format(gr, sel.n_atoms))
                if sel.n_atoms > 0:
                    atomgroups.append(sel)
                else:
                    with warnings.catch_warnings():
                        warnings.simplefilter('always')
                        warnings.warn(
                            "Selection '{}' not taken for profile, "
                            "since it does not contain any atoms.".format(gr))

            if len(atomgroups) == 0:
                raise ValueError("No atoms found in selection."
                                 "Please adjust group selection.")

            if not args._allow_multiple_atomgroups:
                atomgroups = atomgroups[0]

            ana_obj = args.selected_module(atomgroups, verbose=True, save=True)
            # Insert parser arguments into ana_obj
            for var in vars(args):
                if var not in [
                        "topology", "topology_format", "trajectory",
                        "trajectory_format", "atom_style", "begin", "end", "dt",
                        "box", "selection"
                ]:
                    vars(ana_obj)[var] = vars(args)[var]

            ana_obj.run(begin=args.begin, end=args.end, dt=args.dt)

    except Exception as e:
        if args.debug:
            traceback.print_exc()
        else:
            sys.exit("{}Error: {}{}".format(bcolors.fail, e, bcolors.endc))

    if args.debug:
        # Inject local variables into global namespace for debugging.
        for key, value in locals().items():
            globals()[key] = value

        banner = "\nStarting interactive Python interpreter for debug.⁠"
        banner += "\nYou can access all your analysis class atrributes via"
        banner += " ana_obj.<attributes>."
        if use_IPython:
            IPython.embed(banner1=banner)
        else:
            code.interact(banner=banner, local=dict(globals(), **locals()))

    # Only return if ana_obj exists
    if "ana_obj" in locals():
        return ana_obj


def entry_point():
    args = parse_args()
    main(args)


if __name__ == "__main__":
    entry_point()



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import warnings

import numpy as np
from scipy import constants

from .base import MultiGroupAnalysisBase
from ..utils import savetxt, atomgroup_header


def mu(rho, temperature, m):
    """Returns the chemical potential calculated from the density: mu = k_B T log(rho. / m)"""

    # De Broglie (converted to nm)
    db = np.sqrt(constants.h**2 / (2 * np.pi * m * constants.atomic_mass *
                                   constants.Boltzmann * temperature))

    # kT in KJ/mol
    kT = temperature * constants.Boltzmann * constants.Avogadro / constants.kilo

    if np.all(rho > 0):
        return kT * np.log(rho * db**3 / (m * constants.atomic_mass))
    elif np.any(rho == 0):
        return np.float64("-inf")
    else:
        return np.float("nan")


def dmu(rho, drho, temperature):
    """Returns the error of the chemical potential calculated from the density using propagation of uncertainty."""

    if np.all(rho > 0):
        return (drho / rho)
    elif np.any(rho == 0):
        return np.float64("-inf")
    else:
        return np.float("nan")


def weight(selection, dens):
    """Calculates the weights for the histogram depending on the choosen type of density.
        Valid values are `mass`, `number`, `charge` or `temp`."""
    if dens == "mass":
        # amu/nm**3 -> kg/m**3
        return selection.atoms.masses * constants.atomic_mass * 1e27
    elif dens == "number":
        return np.ones(selection.atoms.n_atoms)
    elif dens == "charge":
        return selection.atoms.charges
    elif dens == "temp":
        # ((1 amu * Angstrom^2) / (picoseconds^2)) / Boltzmann constant
        prefac = constants.atomic_mass * 1e4 / constants.Boltzmann
        return ((selection.atoms.velocities**2).sum(axis=1) *
                selection.atoms.masses / 2 * prefac)
    else:
        raise ValueError(
            "`{}` not supported. Use `mass`, `number`, `charge` or `temp`".
            format(dens))


class density_planar(MultiGroupAnalysisBase):
    """Computes partial densities or temperature profiles across the box.

       :param output (str): Output filename
       :param outfreq (int): Default time after which output files are refreshed (1000 ps).
       :param dim (int): Dimension for binning (0=X, 1=Y, 2=Z)
       :param binwidth (float): binwidth (nanometer)
       :param mu (bool): Calculate the chemical potential
       :param muout (str): Prefix for output filename for chemical potential
       :param temperature (float): temperature (K) for chemical potential
       :param mass (float): atommass if not guessed from topology
       :param zpos (float): position at which the chemical potential will be computed. By default average over box.
       :param dens (str): Density: mass, number, charge, electron
       :param comgroup (str): Perform the binning relative to the center of mass of the selected group.
       :param center (bool): Perform the binning relative to the center of the (changing) box.

       :returns (dict): * z: bins
                        * deans_mean: calcualted densities
                        * dens_err: density error
                        * mu: chemical potential
                        * dmu: error of chemical potential
     """

    def __init__(self,
                 atomgroups,
                 output="density.dat",
                 outfreq=1000,
                 dim=2,
                 binwidth=0.1,
                 mu=False,
                 muout="muout.dat",
                 temperature=300,
                 mass=np.nan,
                 zpos=None,
                 dens="mass",
                 comgroup=None,
                 center=False,
                 **kwargs):
        super().__init__(atomgroups, **kwargs)
        self.output = output
        self.outfreq = outfreq
        self.dim = dim
        self.binwidth = binwidth
        self.mu = mu
        self.muout = muout
        self.temperature = temperature
        self.mass = mass
        self.zpos = zpos
        self.dens = dens
        self.comgroup = comgroup
        self.center = center

    def _configure_parser(self, parser):
        parser.add_argument('-o', dest='output')
        parser.add_argument('-dout', dest='outfreq')
        parser.add_argument('-d', dest='dim')
        parser.add_argument('-dz', dest='binwidth')
        parser.add_argument('-mu', dest='mu')
        parser.add_argument('-muo', dest='muout')
        parser.add_argument('-temp', dest='temperature')
        parser.add_argument('-zpos', dest='zpos')
        parser.add_argument('-dens', dest='dens')
        parser.add_argument('-com', dest='comgroup')
        parser.add_argument('-center', dest='center')

    def _prepare(self):
        if self.dens not in ["mass", "number", "charge", "temp"]:
            raise ValueError(
                "Invalid choice for dens: '{}' (choose from 'mass', "
                "'number', 'charge', 'temp')".format(self.dens))
        if self._verbose:
            if self.dens == 'temp':
                print('Computing temperature profile along {}-axes.'.format(
                    'XYZ' [self.dim]))
            else:
                print('Computing {} density profile along {}-axes.'.format(
                    self.dens, 'XYZ' [self.dim]))

        self.ngroups = len(self.atomgroups)
        self.nbins = int(
            np.ceil(self._universe.dimensions[self.dim] / 10 / self.binwidth))

        self.density_mean = np.zeros((self.nbins, self.ngroups))
        self.density_mean_sq = np.zeros((self.nbins, self.ngroups))
        self.av_box_length = 0

        if self.mu and self.dens != 'mass':
            raise ValueError(
                "Calculation of the chemical potential is only possible when "
                "mass density is selected")

        if self.mu:
            if len(self.atomgroups) != 1:
                with warnings.catch_warnings():
                    warnings.simplefilter('always')
                    warnings.warn(
                        "Performing chemical potential analysis for 1st selection"
                        "group '{}'".format(self.atomgroups[0]))
            self.mass = self.atomgroups[0].total_mass(
            ) / self.atomgroups[0].atoms.n_residues

        if self.comgroup is not None:
            self.comsel = self._universe.select_atoms(self.comgroup)
            if self._verbose:
                print("{:>15}: {:>10} atoms".format(self.comgroup,
                                                    self.comsel.n_atoms))
            if self.comsel.n_atoms == 0:
                raise ValueError(
                    "`{}` does not contain any atoms. Please adjust 'com' selection."
                    .format(self.comgroup))
        if self.comgroup is not None:
            self.center = True  # always center when COM
        if self._verbose:
            print("\n")
            print('Using', self.nbins, 'bins.')

    def _single_frame(self):
        curV = self._ts.volume / 1000
        self.av_box_length += self._ts.dimensions[self.dim] / 10
        """ center of mass calculation with generalization to periodic systems
        see Bai, Linge; Breen, David (2008). "Calculating Center of Mass in an
        Unbounded 2D Environment". Journal of Graphics, GPU, and Game Tools. 13
        (4): 53–60. doi:10.1080/2151237X.2008.10129266,
        https://en.wikipedia.org/wiki/Center_of_mass#Systems_with_periodic_boundary_conditions
        """
        if self.comgroup is None:
            comshift = 0
        else:
            Theta = self.comsel.positions[:, self.dim] / self._ts.dimensions[
                self.dim] * 2 * np.pi
            Xi = (np.cos(Theta) *
                  self.comsel.masses).sum() / self.comsel.masses.sum()
            Zeta = (np.sin(Theta) *
                    self.comsel.masses).sum() / self.comsel.masses.sum()
            ThetaCOM = np.arctan2(-Zeta, -Xi) + np.pi
            comshift = self._ts.dimensions[self.dim] * (0.5 - ThetaCOM /
                                                        (2 * np.pi))

        dz = self._ts.dimensions[self.dim] / self.nbins

        for index, selection in enumerate(self.atomgroups):
            bins = (
                (selection.atoms.positions[:, self.dim] + comshift + dz / 2) /
                dz).astype(int) % self.nbins
            density_ts = np.histogram(bins,
                                      bins=np.arange(self.nbins + 1),
                                      weights=weight(selection, self.dens))[0]

            if self.dens == 'temp':
                bincount = np.bincount(bins, minlength=self.nbins)
                self.density_mean[:, index] += density_ts / bincount
                self.density_mean_sq[:, index] += (density_ts / bincount)**2
            else:
                self.density_mean[:, index] += density_ts / curV * self.nbins
                self.density_mean_sq[:, index] += (density_ts / curV *
                                                   self.nbins)**2

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        self._index = self._frame_index + 1

        self.results["dens_mean"] = self.density_mean / self._index
        self.results["dens_mean_sq"] = self.density_mean_sq / self._index

        self.results["dens_std"] = np.nan_to_num(
            np.sqrt(self.results["dens_mean_sq"] -
                    self.results["dens_mean"]**2))
        self.results["dens_err"] = self.results["dens_std"] / \
            np.sqrt(self._index)

        dz = self.av_box_length / (self._index * self.nbins)
        if self.center:
            self.results["z"] = np.linspace(
                -self.av_box_length / self._index / 2,
                self.av_box_length / self._index / 2,
                self.nbins,
                endpoint=False) + dz / 2
        else:
            self.results["z"] = np.linspace(
                0, self.av_box_length / self._index, self.nbins,
                endpoint=False) + dz / 2

        # chemical potential
        if self.mu:
            if (self.zpos is not None):
                this = (self.zpos / (self.av_box_length / self._index) *
                        self.nbins).astype(int)
                if self.center:
                    this += self.nbins // 2
                self.results["mu"] = mu(self.results["dens_mean"][this][0],
                                        self.temperature, self.mass)
                self.results["dmu"] = dmu(self.results["dens_mean"][this][0],
                                          self.results["dens_err"][this][0],
                                          self.temperature)
            else:
                self.results["mu"] = np.mean(
                    mu(self.results["dens_mean"], self.temperature, self.mass))
                self.results["dmu"] = np.mean(
                    dmu(self.results["dens_mean"], self.results["dens_err"],
                        self.temperature))

    def _save_results(self):
        # write header
        if self.dens == "mass":
            units = "kg m^(-3)"
        elif self.dens == "number":
            units = "nm^(-3)"
        elif self.dens == "charge":
            units = "e nm^(-3)"
        elif self.dens == "temp":
            units = "K"

        if self.dens == 'temp':
            columns = "temperature profile [{}]".format(units)
        else:
            columns = "{} density profile [{}]".format(self.dens, units)
        columns += "\nstatistics over {:.1f} picoseconds \npositions [nm]".format(
            self._index * self._universe.trajectory.dt)
        try:
            for group in self.atomgroups:
                columns += "\t" + atomgroup_header(group)
            for group in self.atomgroups:
                columns += "\t" + atomgroup_header(group) + " error"
        except AttributeError:
            with warnings.catch_warnings():
                warnings.simplefilter('always')
                warnings.warn("AtomGroup does not contain resnames."
                              " Not writing residues information to output.")

        # save density profile
        savetxt(self.output,
                np.hstack(
                    ((self.results["z"][:, np.newaxis]),
                     self.results["dens_mean"], self.results["dens_err"])),
                header=columns)

        if self.mu:
            # save chemical potential
            savetxt(self.muout,
                    np.hstack((self.results["mu"], self.results["dmu"]))[None],
                    header="μ [kJ/mol]\t μ error [kJ/mol]")


class density_cylinder(MultiGroupAnalysisBase):
    """Computes partial densities across a cylinder of given radius r and length l.

   :param output (str): Output filename
   :param outfreq (int): Default time after which output files are refreshed (1000 ps).
   :param dim (int): Dimension for binning (0=X, 1=Y, 2=Z)
   :param center (str): Perform the binning relative to the center of this selection string of teh first AtomGroup.
                        If None center of box is used.
   :param radius (float): Radius of the cylinder (nm). If None smallest box extension is taken.
   :param binwidth (float): binwidth (nanometer)
   :param length (float): Length of the cylinder (nm). If None length of box in the binning dimension is taken.
   :param dens (str): Density: mass, number, charge, temp

   :returns (dict): * z: bins
                    * deans_mean: calcualted densities
                    * dens_err: density error
    """

    def __init__(self,
                 atomgroups,
                 output="density_cylinder.dat",
                 outfreq=1000,
                 dim=2,
                 center=None,
                 radius=None,
                 binwidth=0.1,
                 length=None,
                 dens="mass",
                 **kwargs):
        super().__init__(atomgroups, **kwargs)
        self.output = output
        self.outfreq = outfreq
        self.dim = dim
        self.binwidth = binwidth
        self.center = center
        self.radius = radius
        self.length = length
        self.dens = dens

    def _configure_parser(self, parser):
        parser.add_argument('-o', dest='output')
        parser.add_argument('-dout', dest='outfreq')
        parser.add_argument('-d', dest='dim')
        parser.add_argument('-center', dest='center')
        parser.add_argument('-r', dest='radius')
        parser.add_argument('-dr', dest='binwidth')
        parser.add_argument('-l', dest='length')
        parser.add_argument('-dens', dest='dens')

    def _prepare(self):

        if self.dens not in ["mass", "number", "charge", "temp"]:
            raise ValueError(
                "Invalid choice for dens: '{}' (choose from 'mass', "
                "'number', 'charge', 'temp')".format(self.dens))

        if self._verbose:
            if self.dens == 'temp':
                print('Computing temperature profile along {}-axes.'.format(
                    'XYZ' [self.dim]))
            else:
                print(
                    'Computing radial {} density profile along {}-axes.'.format(
                        self.dens, 'XYZ' [self.dim]))

        self.odims = np.roll(np.arange(3), -self.dim)[1:]

        if self.center is None:
            if self._verbose:
                print("No center given --> Take from box dimensions.")
            self.centersel = None
            center = self.atomgroups[0].dimensions[:3] / 2
        else:
            self.centersel = self.atomgroups[0].select_atoms(self.center)
            if len(self.centersel) == 0:
                raise RuntimeError("No atoms found in center selection. "
                                   "Please adjust selection!")
            center = self.centersel.center_of_mass()

        if self._verbose:
            print("Initial center at {}={:.3f} nm and {}={:.3f} nm.".format(
                'XYZ' [self.odims[0]], center[self.odims[0]] / 10,
                'XYZ' [self.odims[1]], center[self.odims[1]] / 10))

        if self.radius is None:
            self.radius = self.atomgroups[0].dimensions[self.odims].min() / 2
            if self._verbose:
                print(
                    "No radius given --> Take smallest box extension (r={:.2f} nm)."
                    .format(self.radius / 10))
        else:
            self.radius /= 10

        if self.length is None:
            self.length = self.atomgroups[0].dimensions[self.dim]
            if self._verbose:
                print("No length given --> Take length in {}.".format(
                    'XYZ' [self.dim]))
        else:
            self.length /= 10

        self.ngroups = len(self.atomgroups)
        self.nbins = int(np.ceil(self.radius / 10 / self.binwidth))

        self.density_mean = np.zeros((self.nbins, self.ngroups))
        self.density_mean_sq = np.zeros((self.nbins, self.ngroups))

        self._dr = np.ones(self.nbins) * self.radius / self.nbins
        self._r_bins = np.arange(self.nbins) * self._dr + self._dr
        self._delta_r_sq = self._r_bins**2 - \
            np.insert(self._r_bins, 0, 0)[0:-1]**2  # r_o^2 - r_i^2

        if self._verbose:
            print("\n")
            print('Using', self.nbins, 'bins.')

    def _single_frame(self):
        # calculater center of cylinder.
        if self.center is None:
            center = self.atomgroups[0].dimensions[:3] / 2
        else:
            center = self.centersel.center_of_mass()

        for index, selection in enumerate(self.atomgroups):

            # select cylinder of the given length and radius
            cut = selection.atoms[np.where(
                np.absolute(selection.atoms.positions[:, self.dim] -
                            center[self.dim]) < self.length / 2)[0]]
            cylinder = cut.atoms[np.where(
                np.linalg.norm((cut.atoms.positions[:, self.odims] -
                                center[self.odims]),
                               axis=1) < self.radius)[0]]

            radial_positions = np.linalg.norm(
                (cylinder.atoms.positions[:, self.odims] - center[self.odims]),
                axis=1)
            bins = np.digitize(radial_positions, self._r_bins)
            density_ts = np.histogram(bins,
                                      bins=np.arange(self.nbins + 1),
                                      weights=weight(cylinder, self.dens))[0]

            if self.dens == 'temp':
                bincount = np.bincount(bins, minlength=self.nbins)
                self.density_mean[:, index] += density_ts / bincount
                self.density_mean_sq[:, index] += (density_ts / bincount)**2
            else:
                self.density_mean[:, index] += density_ts * 1000 / (
                    np.pi * self._delta_r_sq * self.length)
                self.density_mean_sq[:, index] += (
                    density_ts * 1000 /
                    (np.pi * self._delta_r_sq * self.length))**2

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        self._index = self._frame_index + 1

        self.results["r"] = (np.copy(self._r_bins) - self._dr / 2) / 10
        self.results["dens_mean"] = self.density_mean / self._index
        self.results["dens_mean_sq"] = self.density_mean_sq / self._index

        self.results["dens_std"] = np.nan_to_num(
            np.sqrt(self.results["dens_mean_sq"] -
                    self.results["dens_mean"]**2))
        self.results["dens_err"] = self.results["dens_std"] / np.sqrt(
            self._index)

    def _save_results(self):
        # write header
        if self.dens == "mass":
            units = "kg m^(-3)"
        elif self.dens == "number":
            units = "nm^(-3)"
        elif self.dens == "charge":
            units = "e nm^(-3)"
        elif self.dens == "temp":
            units = "K"

        if self.dens == 'temp':
            columns = "temperature profile [{}]".format(units)
        else:
            columns = "{} density profile [{}]".format(self.dens, units)
        columns += "\nstatistics over {:.1f} picoseconds \npositions [nm]".format(
            self._index * self._universe.trajectory.dt)
        for group in self.atomgroups:
            columns += "\t" + atomgroup_header(group)
        for group in self.atomgroups:
            columns += "\t" + atomgroup_header(group) + " error"

        # save density profile
        savetxt(self.output,
                np.hstack(
                    ((self.results["r"][:, np.newaxis]),
                     self.results["dens_mean"], self.results["dens_err"])),
                header=columns)



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import numpy as np

from .utils import _share_path

# Translation of atomnames to types/element
atomtypes = {}
with open(os.path.join(_share_path, "atomtypes.dat")) as f:
    for line in f:
        if line[0] != '#':
            elements = line.split()
            atomtypes[elements[0]] = elements[1]

# Cromer-Mann X-ray scattering factors computed from numerical Hartree-Fock wave functions
# See Acta Cryst. A 24 (1968) p. 321
CM_parameters = {}
with open(os.path.join(_share_path, "sfactor.dat")) as f:
    for line in f:
        if line[0] != '#':
            elements = line.split()
            CM_parameters[elements[0]] = type('CM_parameter', (object,), {})()
            CM_parameters[elements[0]].a = np.array(
                elements[2:6], dtype=np.double)
            CM_parameters[elements[0]].b = np.array(
                elements[6:10], dtype=np.double)
            CM_parameters[elements[0]].c = float(elements[10])



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import numpy as np
from scipy.optimize import curve_fit

from .base import SingleGroupAnalysisBase
from ..utils import check_compound, savetxt


def fitfn(x, alpha, tau1, tau2, pref):
    return pref * (alpha * tau1 * (1 + tau1 / x * (np.exp(-x / tau1) - 1)) +
                   (1 - alpha) * tau2 * (1 + tau2 / x *
                                         (np.exp(-x / tau2) - 1)))


class velocity(SingleGroupAnalysisBase):
    """Mean Velocity analysis.

        Reads in coordinates and velocities from a trajectory and calculates a
        velocity profile along a given axis. The obtained profile is averaged over the 4
        symmetric slab halfs.Error bars are estimated via block averaging as described in [1].

        [1] Hess, B. Determining the shear viscosity of model liquids from molecular
           dynamics simulations. The Journal of Chemical Physics 116, 209-217 (2002).

        :param output_suffix (str): Suffix for output filenames
        :param outfreq (int): Default time after which output files are refreshed (1000 ps).
        :param dim (int): Dimension for position binning (0=X, 1=Y, 2=Z)
        :param vdim (int): Dimension for velocity binning (0=X, 1=Y, 2=Z)
        :param nbins (int): Number of bins.
                            For making use of symmetry must be a multiple of 4.
        :param nblock (int): Maximum number of blocks for block averaging error estimate;
                            1 results in standard error
        :param bpbc (bool): Do not make broken molecules whole again (only works if
                            molecule is smaller than shortest box vector

        :returns (dict): * z: bins [nm]
                         * v: velocity profile [m/s]
                         * ees: velocity error estimate [m/s]
                         * symz: symmetrized bins [nm]
                         * symvel: symmetrized velocity profile [m/s]
                         * symees: symmetrized velocity error estimate [m/s]
     """

    def __init__(self,
                 atomgroups,
                 output_suffix="com",
                 outfreq=1000,
                 dim=2,
                 vdim=0,
                 nbins=200,
                 nblock=10,
                 bpbc=True,
                 **kwargs):
        super().__init__(atomgroups, **kwargs)
        self.output_suffix = output_suffix
        self.outfreq = outfreq
        self.dim = dim
        self.vdim = vdim
        self.nbins = nbins
        self.nblock = nblock
        self.bpbc = bpbc

    def _configure_parser(self, parser):
        parser.add_argument("-o", dest="output_suffix")
        parser.add_argument("-dout", dest="outfreq")
        parser.add_argument("-d", dest="dim")
        parser.add_argument("-dv", dest="vdim")
        parser.add_argument("-nbins", dest="nbins")
        parser.add_argument("-nblock", dest="nblock")
        parser.add_argument("-nopbcrepair", dest="bpbc")

    def _prepare(self):

        if self.nbins % 2 != 0:
            raise ValueError("Number of bins %d can't be divided by 4!")

        self.blockfreq = int(
            np.ceil((self.stopframe - self.startframe) / self.nblock))
        # skip from initial, not end
        self.skipinitialframes = self.n_frames % self.nblock

        self.av_vel = np.zeros((self.nbins, self.nblock))
        self.av_vel_sq = np.zeros((self.nbins))
        # count frame only to velocity if existing
        self.binframes = np.zeros((self.nbins, self.nblock))
        self.L = 0

    def _single_frame(self):
        if self.bpbc:
            # make broken molecules whole again!
            self._universe.atoms.unwrap(compound=check_compound(self._universe.atoms))

        self.L += self._universe.dimensions[self.dim]

        coms = self.atomgroup.center_of_mass(compound=check_compound(self.atomgroup))[:, self.dim]

        comvels = self.atomgroup.atoms.accumulate(
            self.atomgroup.atoms.velocities[:, self.vdim] *
            self.atomgroup.atoms.masses,
            compound=check_compound(self.atomgroup)
        )
        comvels /= self.atomgroup.atoms.accumulate(self.atomgroup.atoms.masses,
                                                   compound=check_compound(self.atomgroup))

        bins = (coms / (self._universe.dimensions[self.dim] / self.nbins)
               ).astype(int) % self.nbins
        bincount = np.bincount(bins, minlength=self.nbins)
        with np.errstate(divide="ignore", invalid="ignore"):
            # mean velocity in this bin, zero if empty
            curvel = np.nan_to_num(
                np.histogram(bins,
                             bins=np.arange(0, self.nbins + 1),
                             weights=comvels)[0] / bincount)

        # add velocities to the average and convert to (m/s)
        self.av_vel[:, self._frame_index // self.blockfreq] += curvel * 100
        self.av_vel_sq[:] += (curvel * 100)**2
        # only average velocities if bin is not empty
        self.binframes[:, self._frame_index // self.blockfreq] += bincount > 0

        if (self._save and self._frame_index % self.outfreq == 0 and
                self._frame_index > 0):
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        """Calculate the results."""

        self._index = self._frame_index + 1

        # minimum number of frames where molecules should be present
        self.minframes = self._index / 100
        avL = self.L / self._index / 10  # in nm
        dz = avL / self.nbins
        self.results["symz"] = np.arange(0, avL / 4 - dz / 2, dz) + dz / 2

        self.results["z"] = np.arange(0, avL - dz / 2, dz) + dz / 2
        self.results["v"] = np.sum(
            self.av_vel[np.sum(self.binframes, axis=1) > self.minframes],
            axis=1) / np.sum(
                self.binframes[np.sum(self.binframes, axis=1) > self.minframes],
                axis=1)
        self.results["dv"] = np.sqrt(self.av_vel_sq[
            np.sum(self.binframes, axis=1) > self.minframes] / np.sum(
                self.binframes[np.sum(self.binframes, axis=1) > self.minframes],
                axis=1) - self.results["v"]**2) / np.sqrt(
                    np.sum(self.binframes[
                        np.sum(self.binframes, axis=1) > self.minframes],
                           axis=1) - 1)

        # make use of the symmetry
        self.results["symvel"] = (
            self.av_vel[:self.nbins // 4] -
            self.av_vel[self.nbins // 4:2 * self.nbins // 4][::-1] -
            self.av_vel[2 * self.nbins // 4:3 * self.nbins // 4] +
            self.av_vel[3 * self.nbins // 4:][::-1])
        self.results["symvel_sq"] = (
            self.av_vel_sq[:self.nbins // 4] +
            self.av_vel_sq[self.nbins // 4:2 * self.nbins // 4][::-1] +
            self.av_vel_sq[2 * self.nbins // 4:3 * self.nbins // 4] +
            self.av_vel_sq[3 * self.nbins // 4:][::-1])
        self.results["symbinframes"] = (
            self.binframes[:self.nbins // 4] +
            self.binframes[self.nbins // 4:2 * self.nbins // 4][::-1] +
            self.binframes[2 * self.nbins // 4:3 * self.nbins // 4] +
            self.binframes[3 * self.nbins // 4:][::-1])

        self.results["vsym"] = np.sum(
            self.results["symvel"][
                np.sum(self.results["symbinframes"], axis=1) > self.minframes],
            axis=1,
        ) / np.sum(
            self.results["symbinframes"][
                np.sum(self.results["symbinframes"], axis=1) > self.minframes],
            axis=1,
        )
        self.results["dvsym"] = np.sqrt(self.results["symvel_sq"][np.sum(
            self.results["symbinframes"], axis=1) > self.minframes] / np.sum(
                self.results["symbinframes"]
                [np.sum(self.results["symbinframes"], axis=1) > self.minframes],
                axis=1,
            ) - self.results["vsym"]**2) / np.sqrt(
                np.sum(
                    self.results["symbinframes"][np.sum(
                        self.results["symbinframes"], axis=1) > self.minframes],
                    axis=1,
                ) - 1)

    def _blockee(self, data):
        ee = []
        for i in range(0, int(np.log2(self.nblock)) - 1):
            bs = 2**i
            numb = self.nblock // bs
            blocks = np.vstack([
                np.mean(data[:, bs * i:bs * (i + 1)], axis=1)
                for i in range(numb)
            ]).T
            ee.append([
                bs * self._trajectory.dt * self.step * self.blockfreq,
                np.std(blocks, axis=1) / np.sqrt(numb - 1),
            ])
        return ee

    def _conclude(self):
        bee = self._blockee(np.nan_to_num(self.av_vel / self.binframes))
        self.results["ee_out"] = np.vstack(
            list(np.hstack((bee[i])) for i in range(len(bee))))

        prefs = (2 * (
            self.av_vel_sq[np.sum(self.binframes, axis=1) > self.minframes] /
            np.sum(
                self.binframes[np.sum(self.binframes, axis=1) > self.minframes],
                axis=1,
            ) - self.results["v"]**2) /
                 (self._index * self._trajectory.dt * self.step)
                )  # 2 sigma^2 / T, (A16) in [1]
        self.results["ees"] = []
        self.results["params"] = []
        for count, i in enumerate(range(self.nbins)):
            if np.sum(self.binframes[i]) > self.minframes:
                pref = prefs[count]

                def modfitfn(x, alpha, tau1, tau2):
                    return fitfn(x, alpha, tau1, tau2, pref)

                [alpha, tau1, tau2], pcov = curve_fit(
                    modfitfn,
                    self.results["ee_out"][:, 0],
                    (self.results["ee_out"][:, i + 1])**2,
                    bounds=([0, 0, 0], [1, np.inf, np.inf]),
                    p0=[0.99, 0.001, 0.01],
                    max_nfev=1e5,
                )
                # (A.17) in [1]
                errest = np.sqrt(pref * (alpha * tau1 + (1 - alpha) * tau2))
                self.results["ees"].append(errest)
                self.results["params"].append([pref, alpha, tau1, tau2])

        # Same for symmetrized
        bee = self._blockee(
            np.nan_to_num(self.results["symvel"] /
                          self.results["symbinframes"]))
        self.results["symee_out"] = np.vstack(
            list(np.hstack((bee[i])) for i in range(len(bee))))

        prefs = (2 * (self.results["symvel_sq"][np.sum(
            self.results["symbinframes"], axis=1) > self.minframes] / np.sum(
                self.results["symbinframes"]
                [np.sum(self.results["symbinframes"], axis=1) > self.minframes],
                axis=1,
            ) - self.results["vsym"]**2) /
                 (self._index * self._trajectory.dt * self.step)
                )  # 2 sigma^2 / T, (A16) in [1]
        self.results["symees"] = []
        for count, i in enumerate(range(self.nbins // 4)):
            if np.sum(self.results["symbinframes"][i]) > self.minframes:
                pref = prefs[count]

                def modfitfn(x, alpha, tau1, tau2):
                    return fitfn(x, alpha, tau1, tau2, pref)

                [alpha, tau1, tau2], pcov = curve_fit(
                    modfitfn,
                    self.results["symee_out"][:, 0],
                    (self.results["symee_out"][:, i + 1])**2,
                    bounds=([0, 0, 0], [1, np.inf, np.inf]),
                    p0=[0.9, 1e3, 1e4],
                    max_nfev=1e5,
                )
                # (A.17) in [1]
                errest = np.sqrt(pref * (alpha * tau1 + (1 - alpha) * tau2))
                self.results["symees"].append(errest)

        if self._save:
            savetxt(
                "errest_" + self.output_suffix,
                np.concatenate(
                    (
                        self.results["ee_out"][:, 0].reshape(
                            len(self.results["ee_out"]), 1),
                        (self.results["ee_out"][:, 1:]
                        )[:, np.sum(self.binframes, axis=1) > self.minframes],
                    ),
                    axis=1,
                ),
                header="z " + " ".join(
                    map(
                        str,
                        self.results["z"][
                            np.sum(self.binframes, axis=1) > self.minframes],
                    )),
            )
            savetxt("errparams_" + self.output_suffix,
                    np.array(self.results["params"]))
            savetxt(
                "errest_sym_" + self.output_suffix,
                np.concatenate(
                    (
                        self.results["symee_out"][:, 0].reshape(
                            len(self.results["symee_out"]), 1),
                        (self.results["symee_out"][:, 1:]
                        )[:,
                          np.sum(self.results["symbinframes"], axis=1) > self.
                          minframes],
                    ),
                    axis=1,
                ),
                header="z " + " ".join(
                    map(
                        str,
                        self.results["symz"]
                        [np.sum(self.results["symbinframes"], axis=1) >
                         self.minframes],
                    )),
            )
            savetxt(
                "errparams_sym_" + self.output_suffix,
                np.array(self.results["params"]),
            )

            savetxt(
                "vel_" + self.output_suffix,
                np.vstack((
                    self.results["z"][
                        np.sum(self.binframes, axis=1) > self.minframes],
                    self.results["v"],
                    np.array(self.results["ees"]),
                    self.results["dv"],
                )).T,
            )

            savetxt(
                "vel_sym_" + self.output_suffix,
                np.vstack((
                    self.results["symz"][np.sum(self.results["symbinframes"],
                                                axis=1) > self.minframes],
                    self.results["vsym"],
                    np.array(self.results["symees"]),
                )).T,
            )

    def _save_results(self):
        savetxt(
            "vel_sym_" + self.output_suffix,
            np.vstack((
                self.results["symz"]
                [np.sum(self.results["symbinframes"], axis=1) > self.minframes],
                self.results["vsym"],
                self.results["dvsym"],
            )).T,
        )
        savetxt(
            "vel_" + self.output_suffix,
            np.vstack((
                self.results["z"][
                    np.sum(self.binframes, axis=1) > self.minframes],
                self.results["v"],
                self.results["dv"],
            )).T,
        )



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import numpy as np

from ..utils import check_compound, savetxt
from .base import SingleGroupAnalysisBase


class dipole_angle(SingleGroupAnalysisBase):
    """Calculates the timeseries of the dipole moment with respect to an axis.

    :param dim (int): refernce vector for angle (x,y,z=0,1,2)
    :param outfreq (float): Default number of frames after which output files are refreshed
    :param output (str): Prefix for output filenames

    :returns (dict): * t: time (ps)
                     * cos_theta_i: Average cos between dipole and axis
                     * cos_theta_ii: Average cos^2 of the same between dipole and axis
                     * cos_theta_ij: Product cos of dipole i and cos of dipole j (i!=j)
    """

    def __init__(self,
                 atomgroup,
                 output="dipangle.dat",
                 outfreq=10000,
                 dim=2,
                 **kwargs):
        super().__init__(atomgroup, **kwargs)
        self.output = output
        self.dim = dim
        self.outfreq = outfreq

    def _configure_parser(self, parser):
        parser.description = self.__doc__
        parser.add_argument('-d', dest='dim')
        parser.add_argument('-dout', dest='outfreq')
        parser.add_argument('-o', dest='output')

    def _prepare(self):
        self.n_residues = self.atomgroup.residues.n_residues

        # unit normal vector
        self.unit = np.zeros(3)
        self.unit[self.dim] += 1

        self.cos_theta_i = np.empty(self.n_frames)
        self.cos_theta_ii = np.empty(self.n_frames)
        self.cos_theta_ij = np.empty(self.n_frames)

    def _single_frame(self):

        # make broken molecules whole again!
        self.atomgroup.unwrap(compound="molecule")

        chargepos = self.positions * self.charges[:, np.newaxis]
        dipoles = self.atomgroup.accumulate(chargepos, compound=check_compound(self.atomgroup))

        cos_theta = np.dot(dipoles, self.unit) / \
            np.linalg.norm(dipoles, axis=1)
        matrix = np.outer(cos_theta, cos_theta)

        trace = matrix.trace()
        self.cos_theta_i[self._frame_index] = cos_theta.mean()
        self.cos_theta_ii[self._frame_index] = trace / self.n_residues
        self.cos_theta_ij[self._frame_index] = (matrix.sum() - trace)
        self.cos_theta_ij[self._frame_index] /= (self.n_residues**2 -
                                                 self.n_residues)

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        self._index = self._frame_index + 1

        self.results["t"] = self._trajectory.dt * \
            np.arange(self.startframe, self.stopframe, self.step)

        self.results["cos_theta_i"] = self.cos_theta_i[:self._index]
        self.results["cos_theta_ii"] = self.cos_theta_ii[:self._index]
        self.results["cos_theta_ij"] = self.cos_theta_ij[:self._index]

    def _save_results(self):

        savetxt(self.output,
                np.vstack([
                    self.results["t"], self.results["cos_theta_i"],
                    self.results["cos_theta_ii"], self.results["cos_theta_ij"]
                ]).T,
                header="t\t<cos(θ_i)>\t<cos(θ_i)cos(θ_i)>\t<cos(θ_i)cos(θ_j)>",
                fmt='%.5e')


class kinetic_energy(SingleGroupAnalysisBase):
    """Calculates the timeseries for the molecular center
       translational and rotational kinetic energy (kJ/mole).

       :param output (str): Output filename
       :param refpoint (str): reference point for molecular center: center of
                              mass (COM), center of charge (COC), or oxygen position (OXY)
                              Note: The oxygen position only works for systems of pure water

        :returns (dict): * t: time (ps)
                         * trans: translational kinetic energy (kJ/mole)
                         * rot: rotational kinetic energy (kJ/mole)
        """

    def __init__(self, atomgroup, output="ke.dat", refpoint="COM", **kwargs):
        super().__init__(atomgroup, **kwargs)
        self.output = output
        self.refpoint = refpoint

    def _configure_parser(self, parser):
        parser.add_argument('-o', dest='output')
        parser.add_argument('-r', dest='refpoint')

    def _prepare(self):
        """Set things up before the analysis loop begins"""
        if self.refpoint not in ["COM", "COC", "OXY"]:
            raise ValueError(
                "Invalid choice for dens: '{}' (choose from 'COM', "
                "'COC', 'OXY')".format(self.refpoint))

        if self.refpoint == "OXY":
            self.oxy = self.atomgroup.select_atoms("name OW*")

        self.masses = self.atomgroup.atoms.accumulate(
            self.atomgroup.atoms.masses, compound=check_compound(self.atomgroup))
        self.abscharges = self.atomgroup.atoms.accumulate(np.abs(
            self.atomgroup.atoms.charges), compound=check_compound(self.atomgroup))
        # Total kinetic energy
        self.E_kin = np.zeros(self.n_frames)

        # Molecular center energy
        self.E_center = np.zeros(self.n_frames)

    def _single_frame(self):
        self.E_kin[self._frame_index] = np.dot(
            self.atomgroup.masses,
            np.linalg.norm(self.atomgroup.velocities, axis=1)**2)

        if self.refpoint == "COM":
            massvel = self.atomgroup.velocities * \
                self.atomgroup.masses[:, np.newaxis]
            v = self.atomgroup.accumulate(massvel, compound=check_compound(self.atomgroup))
            v /= self.masses[:, np.newaxis]

        elif self.refpoint == "COC":
            abschargevel = self.atomgroup.velocities * \
                np.abs(self.atomgroup.charges)[:, np.newaxis]
            v = self.atomgroup.accumulate(abschargevel, compound=check_compound(self.atomgroup))
            v /= self.abscharges[:, np.newaxis]

        elif self.refpoint == "OXY":
            v = self.oxy.velocities

        self.E_center[self._frame_index] = np.dot(self.masses,
                                                  np.linalg.norm(v, axis=1)**2)

    def _calculate_results(self):
        self.results["t"] = self._trajectory.dt * \
            np.arange(self.startframe, self.stopframe, self.step)
        self.results["trans"] = self.E_center / 2 / 100
        self.results["rot"] = (self.E_kin - self.E_center) / 2 / 100

    def _save_results(self):
        savetxt(self.output,
                np.vstack([
                    self.results["t"], self.results["trans"],
                    self.results["rot"]
                ]).T,
                fmt='%.8e',
                header="t / ps \t E_kin^trans / kJ/mole \t E_kin^rot / kJ/mole")



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2020 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import functools
import warnings

import numpy as np


def charge_neutral(filter):
    """Class Decorator to raise an Error/Warning when AtomGroup in an AnalysisBase class
    is not charge neutral. The behaviour of the warning can be controlled
    with the filter attribute. If the AtomGroup's corresponding universe is non-neutral
    an ValueError is raised.

    :param filter (str): Filter type to control warning filter
                         Common values are: "error" or "default"
                         See `warnings.simplefilter` for more options.
    """
    def inner(original_class):
        def charge_check(function):
            @functools.wraps(function)
            def wrapped(self):
                # Check if SingleGroupAnalysis
                if hasattr(self, 'atomgroup'):
                    groups = [self.atomgroup]
                else:
                    groups = self.atomgroups
                for group in groups:
                    if not np.allclose(
                            group.total_charge(compound='fragments'), 0.0):
                        with warnings.catch_warnings():
                            warnings.simplefilter(filter)
                            warnings.warn(
                                "At least one AtomGroup has free charges. "
                                "Analysis for systems with free charges could lead "
                                "to severe artifacts!")

                    if not np.allclose(group.universe.atoms.total_charge(),
                                       0.0):
                        raise ValueError(
                            "Analysis for non-neutral systems is not supported."
                        )
                return function(self)

            return wrapped

        original_class._prepare = charge_check(original_class._prepare)

        return original_class

    return inner



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import re
import sys
import inspect

PARAM_OR_RETURNS_REGEX = re.compile(":(?:param|returns)")
RETURNS_REGEX = re.compile(r":returns \((?P<p_type>.*?)\): (?P<doc>.*)", re.S)
PARAM_REGEX = re.compile(
    r":param (?P<name>[\*\w]+) \((?P<p_type>.*?)\): (?P<doc>.*?)"
    r"(?:(?=:param)|(?=:return)|(?=:raises)|\Z)", re.S)
str_type_dict = {
    "bool": bool,
    "str": str,
    "list": list,
    "int": int,
    "float": float,
    "complex": complex
}


def _trim(docstring):
    """trim function from PEP-257"""
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def _reindent(string):
    return "\n".join(l.strip() for l in string.strip().split("\n"))


def parse_docstring(docstring):
    """Parse the docstring into its components.

    Taken from openstack rally repository.

    :returns: a dictionary of form
              {
                  "short_description": ...,
                  "long_description": ...,
                  "params": [{"name": ..., "doc": ...}, ...],
                  "returns": ...
              }
    """

    short_description = long_description = returns = ""
    params = []

    if docstring:
        docstring = _trim(docstring)

        lines = docstring.split("\n", 1)
        short_description = lines[0]

        if len(lines) > 1:
            long_description = lines[1].strip()

            params_returns_desc = None

            match = PARAM_OR_RETURNS_REGEX.search(long_description)
            if match:
                long_desc_end = match.start()
                params_returns_desc = long_description[long_desc_end:].strip()
                long_description = long_description[:long_desc_end].rstrip()

            if params_returns_desc:
                params = [{
                    "name": name,
                    "type": str_type_dict[p_type],
                    "doc": _trim(doc)
                } for name, p_type, doc in PARAM_REGEX.findall(
                    params_returns_desc)]

                match = RETURNS_REGEX.search(params_returns_desc)
                if match:
                    returns = _reindent(match.group("doc"))

    return {
        "short_description": short_description,
        "long_description": long_description,
        "params": params,
        "returns": returns
    }


def _create_doctsring_dict(func):
    """Creates a dictionary containing arguments and returns including default values from a function"""

    doctsring_dict = parse_docstring(func.__doc__)
    init_arguments = inspect.signature(func).parameters.items()

    # Find default argument
    for function_param, value in init_arguments:
        if value.default is inspect.Parameter.empty:
            continue
        for argument_dict in doctsring_dict["params"]:
            if argument_dict["name"] == function_param:
                argument_dict["default"] = value.default

    return doctsring_dict


def complete_parser(parser, module):
    """Completes options of a parser object based on the values in a docstring dictionary"""

    doctsring_dict = _create_doctsring_dict(module)
    parser.description = doctsring_dict["short_description"] + \
        "\n" + doctsring_dict["long_description"]
    for i, action in enumerate(parser._actions):
        for param_dict in doctsring_dict["params"]:
            if action.dest == param_dict["name"]:

                # Create new action if type is bool
                if param_dict["type"] is bool:

                    option_strings = [s for s in action.option_strings]
                    for s in action.option_strings:
                        parser._handle_conflict_resolve(None, [(s, action)])

                    parser.add_argument(*option_strings,
                                        dest=action.dest,
                                        action='store_' +
                                        str(not param_dict["default"]).lower(),
                                        default=param_dict["default"],
                                        help=param_dict["doc"])

                    parser._actions.insert(i, parser._actions.pop(-1))

                else:  # Add attributes if non boolean parameter
                    action.type = param_dict["type"]
                    action.default = param_dict["default"]
                    action.help = param_dict["doc"]



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import warnings
import sys
import os

from .modules import *

__all__ = [
    'density_planar',
    'density_cylinder',
    'epsilon_bulk',
    'epsilon_planar',
    'epsilon_cylinder',
    'dielectric_spectrum',
    'saxs',
    'debye',
    'dipole_angle',
    'kinetic_energy',
    'diporder',
    'velocity',
]

try:
    sys.path.append(os.path.join(os.path.expanduser("~"),
                                 ".maicos/"))
    from maicos_costum_modules import *
    __all__ += custom_modules
except ImportError:
    pass

__authors__ = "Philip Loche et. al."
__version__ = "0.2"  # NOTE: keep in sync with VERSION in setup.py



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

from .density import *
from .epsilon import *
from .structure import *
from .timeseries import *
from .transport import *



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import sys
import warnings

import numpy as np
from MDAnalysis import NoDataError

_share_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "share")


def repairMolecules(selection):
    """Repairs molecules that are broken due to peridodic boundaries.
    To this end the center of mass is reset into the central box.
    CAVE: Only works with small (< half box) molecules."""

    warnings.warn(
        "repairMolecules is deprecated, use AtomGroup.unwrap() from MDAnalysis instead.",
        category=DeprecationWarning)

    # we repair each moleculetype individually for performance reasons
    for seg in selection.segments:
        atomsPerMolecule = seg.atoms.n_atoms // seg.atoms.n_residues

        # Make molecules whole, use first atom as reference
        distToFirst = np.empty((seg.atoms.positions.shape))
        for i in range(atomsPerMolecule):
            distToFirst[i::atomsPerMolecule] = seg.atoms.positions[i::atomsPerMolecule] - \
                seg.atoms.positions[0::atomsPerMolecule]
        seg.atoms.positions -= (
            np.abs(distToFirst) > selection.dimensions[:3] /
            2.) * selection.dimensions[:3] * np.sign(distToFirst)

        # Calculate the centers of the objects ( i.e. Molecules )
        masspos = (seg.atoms.positions *
                   seg.atoms.masses[:, np.newaxis]).reshape(
                       (seg.atoms.n_atoms // atomsPerMolecule, atomsPerMolecule,
                        3))
        # all molecules should have same mass
        centers = np.sum(masspos.T, axis=1).T / \
            seg.atoms.masses[:atomsPerMolecule].sum()

        # now shift them back into the primary simulation cell
        seg.atoms.positions += np.repeat(
            (centers % selection.dimensions[:3]) - centers,
            atomsPerMolecule,
            axis=0)


def check_compound(AtomGroup):
    """Checks if compound 'molecules' exists. If not it will
    fallback to 'fragments' or 'residues'.
    """
    if hasattr(AtomGroup, "molnums"):
        return "molecules"

    # Do a try-except to catch NoDataError if no bonds
    try:
        hasattr(AtomGroup, "fragments")
        warnings.warn("Cannot use 'molecules'. Falling back to 'fragments'")
        return "fragments"
    except NoDataError:
        warnings.warn("Cannot use 'molecules'. Falling back to 'residues'")
        return "residues"


dt_dk_tolerance = 1e-8  # Max variation from the mean dt or dk that is allowed (~1e-10 suggested)


def FT(t, x, indvar=True):
    """Discrete fast fourier transform.
    Takes the time series and the function as arguments.
    By default, returns the FT and the frequency:\
    setting indvar=False means the function returns only the FT."""
    a, b = np.min(t), np.max(t)
    dt = (t[-1] - t[0]) / float(len(t) - 1)  # timestep
    if (abs((t[1:] - t[:-1] - dt)) > dt_dk_tolerance).any():
        print(np.max(abs(t[1:] - t[:-1])))
        raise RuntimeError("Time series not equally spaced!")
    N = len(t)
    # calculate frequency values for FT
    k = np.fft.fftshift(np.fft.fftfreq(N, d=dt) * 2 * np.pi)
    # calculate FT of data
    xf = np.fft.fftshift(np.fft.fft(x))
    xf2 = xf * (b - a) / N * np.exp(-1j * k * a)
    if indvar:
        return k, xf2
    else:
        return xf2


def iFT(k, xf, indvar=True):
    """Inverse discrete fast fourier transform.
    Takes the frequency series and the function as arguments.
    By default, returns the iFT and the time series:\
    setting indvar=False means the function returns only the iFT."""
    dk = (k[-1] - k[0]) / float(len(k) - 1)  # timestep
    if (abs((k[1:] - k[:-1] - dk)) > dt_dk_tolerance).any():
        print(np.max(abs(k[1:] - k[:-1])))
        raise RuntimeError("Time series not equally spaced!")
    N = len(k)
    x = np.fft.ifftshift(np.fft.ifft(xf))
    t = np.fft.ifftshift(np.fft.fftfreq(N, d=dk)) * 2 * np.pi
    if N % 2 == 0:
        x2 = x * np.exp(-1j * t * N * dk / 2.) * N * dk / (2 * np.pi)
    else:
        x2 = x * np.exp(-1j * t * (N - 1) * dk / 2.) * N * dk / (2 * np.pi)
    if indvar:
        return t, x2
    else:
        return x2


def Correlation(a, b=None, subtract_mean=False):
    """Uses fast fourier transforms to give the correlation function
    of two arrays, or, if only one array is given, the autocorrelation.
    Setting subtract_mean=True causes the mean to be subtracted from the input data."""
    meana = int(subtract_mean) * np.mean(
        a)  # essentially an if statement for subtracting mean
    a2 = np.append(a - meana,
                   np.zeros(2**int(np.ceil((np.log(len(a)) / np.log(2)))) -
                            len(a)))  # round up to a power of 2
    data_a = np.append(a2,
                       np.zeros(len(a2)))  # pad with an equal number of zeros
    fra = np.fft.fft(data_a)  # FT the data
    if b is None:
        sf = np.conj(
            fra
        ) * fra  # take the conj and multiply pointwise if autocorrelation
    else:
        meanb = int(subtract_mean) * np.mean(b)
        b2 = np.append(
            b - meanb,
            np.zeros(2**int(np.ceil((np.log(len(b)) / np.log(2)))) - len(b)))
        data_b = np.append(b2, np.zeros(len(b2)))
        frb = np.fft.fft(data_b)
        sf = np.conj(fra) * frb
    cor = np.real(np.fft.ifft(sf)[:len(a)]) / np.array(range(
        len(a), 0, -1))  # inverse FFT and normalization
    return cor


def ScalarProdCorr(a, b=None, subtract_mean=False):
    """Gives the correlation function of the scalar product of two vector timeseries.
    Arguments should be given in the form a[t, i], where t is the time variable,
    along which the correlation is calculated, and i indexes the vector components."""
    corr = np.zeros(len(a[:, 0]))

    if b is None:
        for i in range(0, len(a[0, :])):
            corr[:] += Correlation(a[:, i], None, subtract_mean)

    else:
        for i in range(0, len(a[0, :])):
            corr[:] += Correlation(a[:, i], b[:, i], subtract_mean)

    return corr


def get_cli_input():
    """Returns a proper fomatted string of the command line input"""
    program_name = os.path.basename(sys.argv[0])
    return "Command line was: {} {}".format(program_name,
                                            " ".join(sys.argv[1:]))


def savetxt(fname, X, header='', fsuffix=".dat", **kwargs):
    """An extension of the numpy savetxt function.
    Adds the command line input to the header and checks for a doubled defined
    filesuffix."""
    header = "{}\n{}".format(get_cli_input(), header)
    fname = "{}{}".format(fname, (not fname.endswith(fsuffix)) * fsuffix)
    np.savetxt(fname, X, header=header, **kwargs)


def atomgroup_header(AtomGroup):
    """Returns a string containing infos about the AtmGroup containing
    the total number of atoms, the including residues and the number of residues.
    Useful for writing output file headers."""

    unq_res, n_unq_res = np.unique(AtomGroup.residues.resnames,
                                   return_counts=True)
    return "{} atom(s): {}".format(
        AtomGroup.n_atoms, ", ".join(
            "{} {}".format(*i) for i in np.vstack([n_unq_res, unq_res]).T))



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import print_function
from setuptools import setup, Extension, find_packages
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler
import os
import sys
import shutil
import tempfile

VERSION = "0.2"  # NOTE: keep in sync with __version__ in maicos.__init__.py
is_release = 'dev' not in VERSION

# Handle cython modules
try:
    # cython has to be >=0.16 <0.28 to support cython.parallel
    from Cython.Build import cythonize
except ImportError:
    if not is_release:
        print("*** package: Cython not found ***")
        print("MAICos requires cython for development builds")
        sys.exit(1)


def get_numpy_include():
    # From MDAnalysis setup.py
    # Obtain the numpy include directory. This logic works across numpy
    # versions.
    # setuptools forgets to unset numpy's setup flag and we get a crippled
    # version of it unless we do it ourselves.
    try:
        # Python 3 renamed the ``__builin__`` module into ``builtins``.
        # Here we import the python 2 or the python 3 version of the module
        # with the python 3 name. This could be done with ``six`` but that
        # module may not be installed at that point.
        import builtins
    except ImportError:
        import __builtin__ as builtins
    builtins.__NUMPY_SETUP__ = False
    try:
        import numpy as np
    except ImportError:
        print('*** package "numpy" not found ***')
        print('MAICos requires a version of NumPy (>=1.13.3), even for setup.')
        print('Please get it from http://numpy.scipy.org/ or install it through '
              'your package manager.')
        sys.exit(-1)
    return np.get_include()


def hasfunction(cc, funcname, include=None, extra_postargs=None):
    # From MDAnalysis setup.py
    tmpdir = tempfile.mkdtemp(prefix='hasfunction-')
    devnull = oldstderr = None
    try:
        try:
            fname = os.path.join(tmpdir, 'funcname.c')
            with open(fname, 'w') as f:
                if include is not None:
                    f.write('#include {0!s}\n'.format(include))
                f.write('int main(void) {\n')
                f.write('    {0!s};\n'.format(funcname))
                f.write('}\n')
            # Redirect stderr to /dev/null to hide any error messages
            # from the compiler.
            # This will have to be changed if we ever have to check
            # for a function on Windows.
            devnull = open('/dev/null', 'w')
            oldstderr = os.dup(sys.stderr.fileno())
            os.dup2(devnull.fileno(), sys.stderr.fileno())
            objects = cc.compile([fname],
                                 output_dir=tmpdir,
                                 extra_postargs=extra_postargs)
            cc.link_executable(objects, os.path.join(tmpdir, "a.out"))
        except Exception:
            return False
        return True
    finally:
        if oldstderr is not None:
            os.dup2(oldstderr, sys.stderr.fileno())
        if devnull is not None:
            devnull.close()
        shutil.rmtree(tmpdir)


def detect_openmp():
    # From MDAnalysis setup.py
    """Does this compiler support OpenMP parallelization?"""
    print("Attempting to autodetect OpenMP support... ", end="")
    compiler = new_compiler()
    customize_compiler(compiler)
    compiler.add_library('gomp')
    include = '<omp.h>'
    extra_postargs = ['-fopenmp']
    hasopenmp = hasfunction(compiler,
                            'omp_get_num_threads()',
                            include=include,
                            extra_postargs=extra_postargs)
    if hasopenmp:
        print("Compiler supports OpenMP")
    else:
        print("Did not detect OpenMP support.")
    return hasopenmp


if __name__ == "__main__":

    # Windows automatically handles math library linking
    # and will not build if we try to specify one
    if os.name == 'nt':
        mathlib = []
    else:
        mathlib = ['m']

    has_openmp = detect_openmp()
    use_cython = not is_release or bool(os.getenv('USE_CYTHON'))
    source_suffix = '.pyx' if use_cython else '.c'

    pre_exts = [Extension("maicos.lib.sfactor", ["maicos/lib/sfactor" + source_suffix],
                          include_dirs=[get_numpy_include()],
                          extra_compile_args=[
                              '-std=c99', '-ffast-math', '-O3', '-funroll-loops'
                          ] + has_openmp * ['-fopenmp'],
                          extra_link_args=has_openmp * ['-fopenmp'],
                          libraries=mathlib)
                ]

    if use_cython:
        extensions = cythonize(pre_exts)
    else:
        extensions = pre_exts
        # Let's check early for missing .c files
        for ext in extensions:
            for source in ext.sources:
                if not (os.path.isfile(source) and
                        os.access(source, os.R_OK)):
                    raise IOError("Source file '{}' not found. This might be "
                                  "caused by a missing Cython install, or a "
                                  "failed/disabled Cython build.".format(source))

    with open("README.md") as summary:
        LONG_DESCRIPTION = summary.read()

    setup(name='maicos',
          packages=find_packages(),
          version=VERSION,
          license='GPL 3',
          description='Analyse molecular dynamics simulations of '
          'interfacial and confined systems.',
          author="Philip Loche et. al.",
          author_email="ploche@physik.fu-berlin.de",
          long_description=LONG_DESCRIPTION,
          long_description_content_type='text/markdown',
          maintainer="Philip Loche",
          maintainer_email="ploche@physik.fu-berlin.de",
          include_package_data=True,
          ext_modules=extensions,
          install_requires=[
              'MDAnalysis>0.19.2',
              'matplotlib>=2.0.0',
              'numpy>=1.10.4',
              'scipy>=0.17',
              'threadpoolctl>=1.1.0',
          ],
          python_requires='>=3.6',
          entry_points={
              'console_scripts': ['maicos = maicos.__main__:entry_point'],
          },
          keywords=[
              'Science',
              'Molecular Dynamics',
              'Confined Systems',
              'MDAnalysis',
            ],
          project_urls={
                'Source Code': 'https://gitlab.com/netzlab/maicos',
            },
          classifiers=[
              'Development Status :: 4 - Beta',
              'Environment :: Console',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
              'Operating System :: POSIX',
              'Operating System :: MacOS :: MacOS X',
              'Operating System :: Microsoft :: Windows ',
              'Programming Language :: Python',
              'Programming Language :: Python :: 3',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
              'Programming Language :: C',
              'Topic :: Scientific/Engineering',
              'Topic :: Scientific/Engineering :: Bio-Informatics',
              'Topic :: Scientific/Engineering :: Chemistry',
              'Topic :: Scientific/Engineering :: Physics',
              'Topic :: Software Development :: Libraries :: Python Modules',
              'Topic :: System :: Shells',
          ],
          zip_safe=False)



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2019 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import warnings

import numpy as np
from MDAnalysis.analysis import base
from MDAnalysis.lib.log import ProgressMeter

logger = logging.getLogger(__name__)


class _AnalysisBase(base.AnalysisBase):
    """Extends the MDAnalysis base class for defining multi frame analysis."""

    def __init__(self, trajectory, verbose=False, save=False, **kwargs):
        """
        Parameters
        ----------
        trajectory : mda.Reader
            A trajectory Reader
        verbose : bool, optional
           Turn on more logging and debugging, default ``False``
        save : bool, optional
           Save results to a file, default ``False``
        """
        super().__init__(self, **kwargs)

        self._trajectory = trajectory
        self._verbose = verbose
        self._save = save
        self.results = {}

    def _setup_frames(self, trajectory, begin=0, end=None, dt=0):
        """
        Pass a Reader object and define the desired iteration pattern
        through the trajectory

        Parameters
        ----------
        trajectory : mda.Reader
            A trajectory Reader
        begin : float, optional
            start time (ps) for evaluation
        end : float, optional
            end time (ps) for evaluation
        dt : float, optional
            time step (ps) to read analysis frame. If `0` take all frames
        """
        self._trajectory = trajectory

        self.begin = begin
        self.end = end
        self.dt = dt

        with warnings.catch_warnings():
            warnings.simplefilter('always')
            if self.begin > trajectory.totaltime:
                raise ValueError("Start ({:.2f} ps) is larer than total time "
                                 "({:.2f} ps).".format(self.begin,
                                                       trajectory.totaltime))
            elif self.begin > 0:
                startframe = int(begin // trajectory.dt)
            else:
                startframe = 0
            if self.end is not None:
                stopframe = int(end // trajectory.dt)
                self.end += 1  # catch also last frame in loops
            else:
                stopframe = None
            if self.dt > 0:
                step = int(dt // trajectory.dt)
            else:
                step = 1

        startframe, stopframe, step = trajectory.check_slice_indices(
            startframe, stopframe, step)
        self.startframe = startframe
        self.stopframe = stopframe
        self.step = step
        self.n_frames = len(range(startframe, stopframe, step))
        interval = int(self.n_frames // 100)
        if interval == 0:
            interval = 1

        verbose = getattr(self, '_verbose', False)
        self._pm = ProgressMeter(self.n_frames if self.n_frames else 1,
                                 interval=interval,
                                 verbose=verbose)

    def _configure_parser(self, parser):
        """Adds parser options using an argparser object"""
        parser.description = self.__doc__

    def _calculate_results(self):
        """Calculate the results"""
        pass

    def _save_results(self):
        """Saves the results you've gatherd to a file."""
        pass

    def run(self, begin=0, end=None, dt=0, verbose=None):
        """Perform the calculation

        Parameters
        ----------
        begin : float, optional
            start time (ps) for evaluation
        end : float, optional
            end time (ps) for evaluation
        dt : float, optional
            time step (ps) to read analysis frame
        verbose : bool, optional
            Turn on verbosity
        """
        logger.info("Choosing frames to analyze")
        # if verbose unchanged, use class default
        verbose = getattr(self, '_verbose',
                          False) if verbose is None else verbose

        self._setup_frames(self._trajectory, begin, end, dt)
        logger.info("Starting preparation")
        self._prepare()
        for i, ts in enumerate(
                self._trajectory[self.startframe:self.stopframe:self.step]):
            self._frame_index = i
            self._ts = ts
            # logger.info("--> Doing frame {} of {}".format(i+1, self.n_frames))
            self._single_frame()
            self._pm.echo(self._frame_index)
        logger.info("Finishing up")
        self._calculate_results()
        self._conclude()
        if self._save:
            self._save_results()
        return self


class SingleGroupAnalysisBase(_AnalysisBase):
    """The base class for analysing a single AtomGroup only."""

    _allow_multiple_atomgroups = False

    def __init__(self, atomgroup, **kwargs):
        super().__init__(atomgroup.universe.trajectory, **kwargs)
        self.atomgroup = atomgroup
        self._universe = atomgroup.universe


class MultiGroupAnalysisBase(_AnalysisBase):
    """The base class for analysing a single or multiple AtomGroups."""

    _allow_multiple_atomgroups = True

    def __init__(self, atomgroups, **kwargs):
        if type(atomgroups) not in [list, tuple, np.ndarray]:
            atomgroups = [atomgroups]
        else:
            # Check that all atomgroups are from same universe
            for ag in atomgroups[:1]:
                if ag.universe != atomgroups[0].universe:
                    raise ValueError(
                        "Given Atomgroups are not from the same Universe.")
        super().__init__(atomgroups[0].universe.trajectory, **kwargs)

        self.atomgroups = atomgroups
        self._universe = atomgroups[0].universe



#!/usr/bin/env python3
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
#
# Copyright (c) 2020 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
import tempfile

import numpy as np
import MDAnalysis as mda

from .base import SingleGroupAnalysisBase
from .. import tables
from ..lib import sfactor
from ..utils import check_compound, savetxt


def compute_form_factor(q, atom_type):
    """Calculates the form factor for the given element for given q (1/nm).
       Handles united atom types like CH4 etc ..."""
    element = tables.atomtypes[atom_type]

    if element == "CH1":
        form_factor = compute_form_factor(q, "C") + compute_form_factor(q, "H")
    elif element == "CH2":
        form_factor = compute_form_factor(
            q, "C") + 2 * compute_form_factor(q, "H")
    elif element == "CH3":
        form_factor = compute_form_factor(
            q, "C") + 3 * compute_form_factor(q, "H")
    elif element == "CH4":
        form_factor = compute_form_factor(
            q, "C") + 4 * compute_form_factor(q, "H")
    elif element == "NH1":
        form_factor = compute_form_factor(q, "N") + compute_form_factor(q, "H")
    elif element == "NH2":
        form_factor = compute_form_factor(
            q, "N") + 2 * compute_form_factor(q, "H")
    elif element == "NH3":
        form_factor = compute_form_factor(
            q, "N") + 3 * compute_form_factor(q, "H")
    else:
        form_factor = tables.CM_parameters[element].c
        # factor of 10 to convert from 1/nm to 1/Angstroms
        q2 = (q / (4 * np.pi * 10))**2
        for i in range(4):
            form_factor += tables.CM_parameters[element].a[i] * \
                np.exp(-tables.CM_parameters[element].b[i] * q2)

    return form_factor


class saxs(SingleGroupAnalysisBase):
    """Computes SAXS scattering intensities S(q) for all atom types from the given trajectory.

    The q vectors are binned
    by their length using a binwidth given by -dq. Using the -nobin option
    the raw intensity for each q_{i,j,k} vector
    is saved using. Note that this only works reliable using constant box vectors!
    The possible scattering vectors q can be restricted by a miminal and maximal angle with the z-axis.
    For 0 and 180 all possible vectors are taken into account.
    For the scattering factor the structure fator is multiplied by a atom type specific form factor
    based on Cromer-Mann parameters. By using the -sel option atoms can be selected for which the
    profile is calculated. The selection uses the MDAnalysis selection commands.

    :param outfreq (float): Number of frames after which the output is updated.
    :param output (str): Output filename
    :param noboindata (bool): Do not bin the data. Only works reliable for NVT!
    :param startq (float): Starting q (1/nm)
    :param endq (float): Ending q (1/nm)
    :param dq (float): binwidth (1/nm)
    :param mintheta (float): Minimal angle (°) between the q vectors and the z-axis.
    :param maxtheta (float): Maximal angle (°) between the q vectors and the z-axis.

    :returns (dict): * q: length of binned q-vectors
                     * q_indices: Miller indices of q-vector (only if noboindata==True)
                     * scat_factor: Scattering intensities
    """

    def __init__(self,
                 atomgroup,
                 outfreq=100,
                 output="sq.dat",
                 nobin=False,
                 startq=0,
                 endq=60,
                 dq=0.05,
                 mintheta=0,
                 maxtheta=180,
                 **kwargs):
        super().__init__(atomgroup, **kwargs)
        self.outfreq = outfreq
        self.output = output
        self.nobindata = nobin
        self.startq = startq
        self.endq = endq
        self.dq = dq
        self.mintheta = mintheta
        self.maxtheta = maxtheta

    def _configure_parser(self, parser):
        parser.add_argument('-dout', dest='outfreq')
        parser.add_argument('-sq', dest='output')
        parser.add_argument('-startq', dest='startq')
        parser.add_argument('-endq', dest='endq')
        parser.add_argument('-nobin', dest='nobindata')
        parser.add_argument('-dq', dest='dq', type=float, default=0.05, help='')
        parser.add_argument('-mintheta', dest='mintheta')
        parser.add_argument('-maxtheta', dest='maxtheta')

    def _prepare(self):

        self.mintheta = min(self.mintheta, self.maxtheta)
        self.maxtheta = max(self.mintheta, self.maxtheta)

        if self.mintheta < 0 and self._verbose:
            print("mintheta = {}° < 0°: Set mininmal angle to 0°.".format(
                self.mintheta))
            self.mintheta = 0
        if self.maxtheta > 180 and self._verbose:
            print("maxtheta = {}° > 180°: Set maximal angle to 180°.".format(
                self.maxtheta))
            self.maxtheta = np.pi

        self.mintheta *= np.pi / 180
        self.maxtheta *= np.pi / 180

        self.groups = []
        self.atom_types = []
        if self._verbose:
            print("\nMap the following atomtypes:")
        for atom_type in np.unique(self.atomgroup.types).astype(str):
            try:
                element = tables.atomtypes[atom_type]
            except KeyError:
                raise RuntimeError(
                    "No suitable element for '{0}' found. You can add '{0}' together with a suitable element to 'share/atomtypes.dat'."
                    .format(atom_type))
            if element == "DUM":
                continue
            self.groups.append(
                self.atomgroup.select_atoms("type {}*".format(atom_type)))
            self.atom_types.append(atom_type)

            if self._verbose:
                print("{:>14} --> {:>5}".format(atom_type, element))

        if self._verbose:
            print("")

        if self.nobindata:
            self.box = np.diag(
                mda.lib.mdamath.triclinic_vectors(
                    self._universe.dimensions)) / 10
            self.q_factor = 2 * np.pi / self.box
            self.maxn = np.ceil(self.endq / self.q_factor).astype(int)
            self.S_array = np.zeros(list(self.maxn) + [len(self.groups)])
        else:
            self.nbins = int(np.ceil((self.endq - self.startq) / self.dq))
            self.struct_factor = np.zeros([self.nbins, len(self.groups)])

    def _single_frame(self):
        # convert everything to cartesian coordinates
        box = np.diag(mda.lib.mdamath.triclinic_vectors(self._ts.dimensions))
        for i, t in enumerate(self.groups):
            positions = t.atoms.positions - box * \
                np.round(t.atoms.positions / box)  # minimum image

            q_ts, S_ts = sfactor.compute_structure_factor(
                np.double(positions / 10), np.double(box / 10), self.startq,
                self.endq, self.mintheta, self.maxtheta)

            S_ts *= compute_form_factor(q_ts, self.atom_types[i])**2

            if self.nobindata:
                self.S_array[:, :, :, i] += S_ts
            else:
                q_ts = q_ts.flatten()
                S_ts = S_ts.flatten()
                nonzeros = np.where(S_ts != 0)[0]

                q_ts = q_ts[nonzeros]
                S_ts = S_ts[nonzeros]

                bins = ((q_ts - self.startq) /
                        ((self.endq - self.startq) / self.nbins)).astype(int)
                struct_ts = np.histogram(bins,
                                         bins=np.arange(self.nbins + 1),
                                         weights=S_ts)[0]
                with np.errstate(divide='ignore', invalid='ignore'):
                    struct_ts /= np.histogram(bins,
                                              bins=np.arange(self.nbins + 1))[0]
                self.struct_factor[:, i] += np.nan_to_num(struct_ts)

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        self._index = self._frame_index + 1
        if self.nobindata:
            self.results["scat_factor"] = self.S_array.sum(axis=3)
            self.results["q_indices"] = np.array(
                list(np.ndindex(tuple(self.maxn))))
            self.results["q"] = np.linalg.norm(self.results["q_indices"] *
                                               self.q_factor[np.newaxis, :],
                                               axis=1)
        else:
            q = np.arange(self.startq, self.endq, self.dq) + 0.5 * self.dq
            nonzeros = np.where(self.struct_factor[:, 0] != 0)[0]
            scat_factor = self.struct_factor[nonzeros]

            self.results["q"] = q[nonzeros]
            self.results["scat_factor"] = scat_factor.sum(axis=1)

        self.results["scat_factor"] /= (self._index * self.atomgroup.n_atoms)

    def _save_results(self):
        """Saves the current profiles to a file."""

        if self.nobindata:
            out = np.hstack([
                self.results["q"][:, np.newaxis], self.results["q_indices"],
                self.results["scat_factor"].flatten()[:, np.newaxis]
            ])
            nonzeros = np.where(out[:, 4] != 0)[0]
            out = out[nonzeros]
            selfort = np.selfort(out[:, 0])
            out = out[selfort]

            boxinfo = "box_x = {0:.3f} nm, box_y = {1:.3f} nm, box_z = {2:.3f} nm\n".format(
                *self.box)
            savetxt(self.output,
                    out,
                    header=boxinfo +
                    "q (1/nm)\tq_i\t q_j \t q_k \tS(q) (arb. units)",
                    fmt='%.4e')
        else:
            savetxt(self.output,
                    np.vstack([self.results["q"],
                               self.results["scat_factor"]]).T,
                    header="q (1/nm)\tS(q) (arb. units)",
                    fmt='%.4e')


class debye(SingleGroupAnalysisBase):
    """Calculates scattering intensities using the debye equation.

       :param outfreq (float): Number of frames after which the output is updated.
       :param output (str): Output filename
       :param startq (float): Starting q (1/nm)
       :param endq (float): Ending q (1/nm)
       :param dq (float): binwidth (1/nm)
       :param sinc (bool): Apply sinc damping
       :param debyer (str): Path to the debyer executable

       :returns (dict): * q: length of binned q-vectors
                        * scat_factor: Scattering intensities
    """

    def __init__(self,
                 atomgroup,
                 sel="all",
                 outfreq=100,
                 output="sq.dat",
                 startq=0,
                 endq=60,
                 dq=0.05,
                 sinc=False,
                 debyer="debyer",
                 **kwargs):
        super().__init__(atomgroup, **kwargs)
        self.outfreq = outfreq
        self.output = output
        self.startq = startq
        self.endq = endq
        self.dq = dq
        self.sinc = sinc
        self.debyer = debyer

    def _configure_parser(self, parser):
        parser.add_argument('-dout', dest='outfreq')
        parser.add_argument('-sq', dest='output')
        parser.add_argument('-startq', dest='startq')
        parser.add_argument('-endq', dest='endq')
        parser.add_argument(
            '-dq',
            dest='dq',
        )
        parser.add_argument('-sinc', dest='sinc')
        parser.add_argument('-d', dest='debyer')

    def _prepare(self):

        # Convert 1/nm to 1/Å
        self.startq /= 10
        self.endq /= 10
        self.dq /= 10

        self.atomgroup = self.atomgroup.select_atoms(
            "not name DUM and not name MW")

        # Create an extra list for the atom names.
        # This is necessary since it is not possible to efficently add axtra atoms to
        # a MDAnalysis universe, necessary for the hydrogens in united atom forcefields.

        self.atom_names = self.atomgroup.n_atoms * ['']

        for i, atom_type in enumerate(self.atomgroup.types.astype(str)):
            element = tables.atomtypes[atom_type]

            # add hydrogens in the case of united atom forcefields
            if element in ["CH1", "CH2", "CH3", "CH4", "NH", "NH2", "NH3"]:
                self.atom_names[i] = element[0]
                for h in range(int(element[-1])):
                    self.atom_names.append("H")
                    # add a extra atom to universe. It got the wrong type but we only
                    # need the position, since we maintain our own atom type list.
                    self.atomgroup += self.atomgroup.atoms[i]
            else:
                self.atom_names[i] = element

        # create tmp directory for saving datafiles
        self._tmp = tempfile.mkdtemp()

        self._OUT = open(os.devnull, 'w')

        try:
            subprocess.call(self.debyer, stdout=self._OUT, stderr=self._OUT)
        except FileNotFoundError:
            raise RuntimeError("{}: command not found".format(self.debyer))

        if self._verbose:
            print("{} is the tempory directory for all files.\n".format(
                self._tmp))

    def _writeXYZ(self, filename):
        """Writes the positions of the current frame to the given xyz file"""
        write = mda.coordinates.XYZ.XYZWriter(filename,
                                              n_atoms=len(self.atom_names),
                                              atoms=self.atom_names)

        ts = self._trajectory.ts.copy_slice(self.atomgroup.indices)
        write.write_next_timestep(ts)
        write.close()

    def _single_frame(self):

        # convert coordinates in a rectengular box
        box = np.diag(mda.lib.mdamath.triclinic_vectors(self._ts.dimensions))
        self.atomgroup.positions = self.atomgroup.positions \
            - box * np.round(self.atomgroup.positions /
                             box)  # minimum image

        self._writeXYZ("{}/{}.xyz".format(self._tmp, self._frame_index))

        ref_q = 4 * np.pi / np.min(box)
        if ref_q > self.startq:
            self.startq = ref_q

        command = "-x -f {0} -t {1} -s {2} -o {3}/{4}.dat -a {5} -b {6} -c {7} -r {8} {3}/{4}.xyz".format(
            round(self.startq, 3), self.endq, self.dq, self._tmp,
            self._frame_index, box[0], box[1], box[2],
            np.min(box) / 2.001)

        command += self.sinc * " --sinc"

        subprocess.call("{} {}".format(self.debyer, command),
                        stdout=self._OUT,
                        stderr=self._OUT,
                        shell=True)

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        datfiles = [f for f in os.listdir(self._tmp) if f.endswith(".dat")]

        s_tmp = np.loadtxt("{}/{}".format(self._tmp, datfiles[0]))
        for f in datfiles[1:]:
            s_tmp = np.vstack([s_tmp, np.loadtxt("{}/{}".format(self._tmp, f))])

        nbins = int(np.ceil((self.endq - self.startq) / self.dq))
        q = np.arange(self.startq, self.endq, self.dq) + 0.5 * self.dq

        bins = ((s_tmp[:, 0] - self.startq) /
                ((self.endq - self.startq) / nbins)).astype(int)
        s_out = np.histogram(bins,
                             bins=np.arange(nbins + 1),
                             weights=s_tmp[:, 1])[0]

        nonzeros = np.where(s_out != 0)[0]

        self.results["q"] = 10 * q[nonzeros]
        self.results["scat_factor"] = s_out[nonzeros] / len(datfiles)

    def _conclude(self):
        for f in os.listdir(self._tmp):
            os.remove("{}/{}".format(self._tmp, f))

        os.rmdir(self._tmp)

    def _save_results(self):
        savetxt(self.output,
                np.vstack([self.results["q"], self.results["scat_factor"]]).T,
                header="q (1/A)\tS(q)_tot (arb. units)",
                fmt='%.8e')


class diporder(SingleGroupAnalysisBase):
    """Calculates dipolar order parameters including
    the projected polarization density P_0⋅ρ(z)⋅cos(θ[z]),
    the dipole orientation cos(θ[z]),
    the squared dipole orientation cos²(Θ[z]) and
    the number density ρ(z)

    :param binwidth (float): specify the binwidth [nm]
    :param dim (int): direction normal to the surface (x,y,z=0,1,2)
    :param output (str): Output filename
    :param outfreq (int): Default number of frames after which output files are
                         refreshed
    :param bsym (bool): symmetrize the profiles
    :param center (bool): Shift system by half a box length (useful for
                          membrane simulations
    :param com (bool): shift system such that the water COM is centered
    :param binmethod (str): binning method: center of Mass (COM), center of charge (COC)
                            or oxygen position (OXY).
                            Note: `OXY` only works for water oxygens with name `OW*`
    :param bpbc (bool): do not make broken molecules whole again
                       (only works if molecule is smaller than shortest box vector


   :returns (dict): * z: bins [nm]
                    * P0: P_0⋅cos(θ[z]) [e/nm²]
                    * cos_theta: cos(θ[z])
                    * cos_2_theta: cos²(Θ[z])
                    * rho: ρ(z) [1/nm³]
    """

    def __init__(self,
                 atomgroup,
                 binwidth=0.01,
                 dim=2,
                 output="diporder.dat",
                 outfreq=10000,
                 bsym=False,
                 center=False,
                 com=False,
                 binmethod='COM',
                 bpbc=True,
                 **kwargs):
        super().__init__(atomgroup, **kwargs)

        self.binwidth = binwidth
        self.dim = dim
        self.output = output
        self.outfreq = outfreq
        self.bsym = bsym
        self.center = center
        self.com = com
        self.binmethod = binmethod
        self.bpbc = bpbc

    def _configure_parser(self, parser):
        parser.add_argument('-dz', dest='binwidth')
        parser.add_argument('-d', dest='dim')
        parser.add_argument('-o', dest='output')
        parser.add_argument('-dout', dest="outfreq")
        parser.add_argument('-sym', dest='bsym')
        parser.add_argument('-shift', dest='center')
        parser.add_argument('-com', dest='com')
        parser.add_argument('-bin', dest='binmethod')
        parser.add_argument('-nopbcrepair', dest='bpbc')

    def _prepare(self):
        """Set things up before the analysis loop begins"""

        self.binmethod = self.binmethod.upper()
        if self.binmethod not in ["COM", "COC", "OXY"]:
            raise ValueError('Unknown binning method: {}'.format(
                self.binmethod))

        if self.binmethod == "OXY":
            self.oxy = self.atomgroup.select_atoms("name OW*")

        # Check if all residues are identical. Choose first residue as reference.
        residue_names = [ag.names for ag in self.atomgroup.split("residue")]
        for names in residue_names[1:]:
            if len(residue_names[0]) != len(names) and np.all(
                    residue_names[0] != names):
                raise ValueError("Not all residues are identical. Please adjust"
                                 "selection.")

        # Assume a threedimensional universe...
        self.xydims = np.roll(np.arange(3), -self.dim)[1:]
        self.nbins = int(
            np.ceil(self._universe.dimensions[self.dim] / 10 / self.binwidth))
        self.P0 = np.zeros(self.nbins)
        self.cos_theta = np.zeros(self.nbins)
        self.cos_2_theta = np.zeros(self.nbins)
        self.rho = np.zeros(self.nbins)
        self.av_box_length = 0

        # unit normal vector
        self.unit = np.zeros(3)
        self.unit[self.dim] += 1

    def _single_frame(self):

        if self.center:
            # shift membrane
            self._ts.positions[:, self.dim] += self._ts.dimensions[self.dim] / 2
            self._ts.positions[:, self.dim] %= self._ts.dimensions[self.dim]
        if self.com:
            # put water COM into center
            waterCOM = np.sum(self.atomgroup.atoms.positions[:, 2] *
                              self.atomgroup.atoms.masses
                             ) / self.atomgroup.atoms.masses.sum()
            self._ts.positions[:, self.dim] += self._ts.dimensions[
                self.dim] / 2 - waterCOM
            self._ts.positions[:, self.dim] %= self._ts.dimensions[self.dim]

        # Make molecules whole
        if self.bpbc:
            self.atomgroup.unwrap(compound=check_compound(self.atomgroup))

        dz_frame = self._ts.dimensions[self.dim] / self.nbins

        chargepos = self.atomgroup.positions * self.atomgroup.charges[:, np.
                                                                      newaxis]
        dipoles = self.atomgroup.accumulate(chargepos, 
                                            compound=check_compound(self.atomgroup))
        dipoles /= 10  # convert to e nm

        if self.binmethod == 'COM':
            # Calculate the centers of the objects ( i.e. Molecules )
            coms = self.atomgroup.center_of_mass(compound=check_compound(self.atomgroup))
            bins = ((coms[:, self.dim] % self._ts.dimensions[self.dim]) /
                    dz_frame).astype(int)
        elif self.binmethod == 'COC':
            cocs = self.atomgroup.center(weights=np.abs(self.atomgroup.charges),
                                         compound=check_compound(self.atomgroup))
            bins = ((cocs[:, self.dim] % self._ts.dimensions[self.dim]) /
                    dz_frame).astype(int)
        elif self.binmethod == 'OXY':
            bins = ((self.oxy.positions[:, self.dim] %
                     self._ts.dimensions[self.dim]) / dz_frame).astype(int)

        bincount = np.bincount(bins, minlength=self.nbins)
        A = np.prod(self._ts.dimensions[self.xydims])

        self.P0 += np.histogram(
            bins,
            bins=np.arange(self.nbins + 1),
            weights=np.dot(dipoles, self.unit))[0] / (A * dz_frame / 1e3)
        cos_theta = np.histogram(
            bins,
            bins=np.arange(self.nbins + 1),
            weights=np.dot(
                dipoles / np.linalg.norm(dipoles, axis=1)[:, np.newaxis],
                self.unit))[0]
        with np.errstate(divide='ignore', invalid='ignore'):
            self.cos_theta += np.nan_to_num(cos_theta / bincount)
            self.cos_2_theta += np.nan_to_num(cos_theta**2 / bincount)
        self.rho += bincount / (A * dz_frame / 1e3)  # convert to 1/nm^3

        self.av_box_length += self._ts.dimensions[self.dim] / 10

        if self._save and self._frame_index % self.outfreq == 0 and self._frame_index > 0:
            self._calculate_results()
            self._save_results()

    def _calculate_results(self):
        """Calculate the results.

        Called at the end of the run() method to before the _conclude function.
        Can also called during a run to update the results during processing."""
        self._index = self._frame_index + 1

        dz = self.av_box_length / (self._index * self.nbins)
        if self.center:
            self.results["z"] = np.linspace(
                -self.av_box_length / self._index / 2,
                self.av_box_length / self._index / 2,
                self.nbins,
                endpoint=False) + dz / 2
        else:
            self.results["z"] = np.linspace(
                0, self.av_box_length / self._index, self.nbins,
                endpoint=False) + dz / 2

        self.results["P0"] = self.P0 / self._frame_index
        self.results["cos_theta"] = self.cos_theta / self._frame_index
        self.results["cos_2_theta"] = self.cos_2_theta / self._frame_index
        self.results["rho"] = self.rho / self._frame_index

        if self.bsym:
            for i in range(len(self.results["z"]) - 1):
                self.results["z"][i +
                                  1] = .5 * (self.results["z"][i + 1] +
                                             self.results["z"][i + 1][-1::-1])
                self.results["diporder"][
                    i + 1] = .5 * (self.results["diporder"][i + 1] +
                                   self.results["diporder"][i + 1][-1::-1])

    def _save_results(self):
        """Saves results to a file.

        Called at the end of the run() method after _calculate_results and
        _conclude"""

        header = "z [nm]\t"
        header += "P_0*cos(Theta[z]) [e/nm^2]\t"
        header += "cos(theta(z))\t"
        header += "cos^2(theta(z))\t"
        header += "rho(z) [1/nm^3]"

        savetxt(self.output,
                np.vstack([
                    self.results["z"], self.results["P0"],
                    self.results["cos_theta"], self.results["cos_2_theta"],
                    self.results["rho"]
                ]).T,
                header=header)



# distutils: language = c
# cython: language_level=3
#
# Copyright (c) 2020 Authors and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
# SPDX-License-Identifier: GPL-2.0-or-later

import numpy as np

cimport numpy as np
cimport cython
from libc cimport math
from cython.parallel cimport prange

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision
cpdef tuple compute_structure_factor(double[:,:] positions, double[:] boxdimensions,
                                     double start_q, double end_q, double mintheta,
                                     double maxtheta):
    """Calculates S(|q|) for all possible q values. Returns the q values as well as the scattering factor."""

    assert(boxdimensions.shape[0]==3)
    assert(positions.shape[1]==3)

    cdef Py_ssize_t i, j, k, l, n_atoms
    cdef int[::1] maxn = np.empty(3,dtype=np.int32)
    cdef double qx, qy, qz, qrr, qdotr, sin, cos, theta
    cdef double[::1] q_factor = np.empty(3,dtype=np.double)

    n_atoms = positions.shape[0]
    for i in range(3):
        q_factor[i] = 2*np.pi/boxdimensions[i]
        maxn[i] = <int>math.ceil(end_q/<float>q_factor[i])

    cdef double[:,:,::1] S_array = np.zeros(maxn, dtype=np.double)
    cdef double[:,:,::1] q_array = np.zeros(maxn, dtype=np.double)

    for i in prange(<int>maxn[0],nogil=True):
        qx = i * q_factor[0]

        for j in range(maxn[1]):
            qy = j * q_factor[1]

            for k in range(maxn[2]):
                if (i + j + k != 0):
                    qz = k * q_factor[2]
                    qrr = math.sqrt(qx*qx+qy*qy+qz*qz)
                    theta = math.acos(qz / qrr)

                    if (qrr >= start_q and qrr <= end_q and
                          theta >= mintheta and theta <= maxtheta):
                        q_array[i,j,k] = qrr

                        sin = 0
                        cos = 0
                        for l in range(n_atoms):
                            qdotr = positions[l,0]*qx + positions[l,1]*qy + positions[l,2]*qz
                            sin += math.sin(qdotr)
                            cos += math.cos(qdotr)

                        S_array[i,j,k] += sin*sin + cos*cos

    return (np.asarray(q_array), np.asarray(S_array))
